/*
 * MVC controller
 *
 * Copyright(c) 2008 olyutorskii
 * $Id: Controller.java 237 2008-10-09 16:20:22Z olyutorskii $
 */

package jp.sourceforge.jindolf;

import java.awt.Container;
import java.awt.Font;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.font.FontRenderContext;
import java.io.IOException;
import java.net.URL;
import java.util.List;
import java.util.concurrent.Executor;
import java.util.concurrent.Executors;
import java.util.logging.Handler;
import java.util.logging.Level;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JEditorPane;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.LookAndFeel;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.event.TreeWillExpandListener;
import javax.swing.tree.TreePath;

/**
 * いわゆるMVCでいうとこのコントローラ
 */
public class Controller
        implements ActionListener,
                   TreeWillExpandListener,
                   TreeSelectionListener,
                   ChangeListener,
                   AnchorHitListener {
    
    private static final String HELP_HTML = "resources/help.html";

    private ActionManager actionManager;
    private TopFrameView topFrame;
    private LandsModel model;
    private JFrame helpFrame;
    private FilterPanel filterFrame;
    private AccountPanel accountFrame;
    private LogFrame showlogFrame;
    private FontChooser fontChooser;
    private FindPanel findPanel;
    private DaySummary daySummaryPanel;
    private volatile boolean isBusyNow;

    /**
     * コントローラの生成。
     * @param actionManager アクション管理
     * @param topFrame 最上位ビュー
     * @param model 最上位データモデル
     */
    public Controller(ActionManager actionManager,
                       TopFrameView topFrame,
                       LandsModel model){
        super();

        this.actionManager = actionManager;
        this.topFrame = topFrame;
        this.model = model;

        this.actionManager.addActionListener(this);

        JTree treeView = this.topFrame.getTreeView();
        treeView.setModel(this.model);
        treeView.addTreeWillExpandListener(this);
        treeView.addTreeSelectionListener(this);

        this.topFrame.getTabBrowser().addChangeListener(this);
        this.topFrame.getTabBrowser().addActionListener(this);
        this.topFrame.getTabBrowser().addAnchorHitListener(this);
        
        JButton reloadVillageListButton = this.topFrame
                                         .getLandsTree()
                                         .getReloadVillageListButton();
        reloadVillageListButton.addActionListener(this);
        reloadVillageListButton.setEnabled(false);

        this.filterFrame = new FilterPanel(this.topFrame);
        this.filterFrame.addChangeListener(this);
        this.filterFrame.pack();
        this.filterFrame.setVisible(false);
        
        this.showlogFrame = new LogFrame(this.topFrame);
        this.showlogFrame.pack();
        this.showlogFrame.setSize(600,250);
        this.showlogFrame.setLocationByPlatform(true);
        this.showlogFrame.setVisible(false);
        Handler newHandler = this.showlogFrame.getHandler();
        Jindolf.logger.addHandler(newHandler);
        Handler[] handlers = Jindolf.logger.getHandlers();
        for(Handler handler : handlers){
            if( ! (handler instanceof PileHandler) ) continue;
            PileHandler pile = (PileHandler) handler;
            pile.delegate(newHandler);
            pile.close();
        }
        
        this.fontChooser = new FontChooser(this.topFrame);
        this.fontChooser.pack();
        this.fontChooser.setSize(600,500);
        this.fontChooser.setVisible(false);
        initFontChooser();
        
        this.topFrame
            .getTabBrowser()
            .setFontInfo(this.fontChooser.getSelectedFont(),
                         this.fontChooser.getFontRenderContext());
        
        this.findPanel = new FindPanel(this.topFrame);
        this.findPanel.pack();
        this.findPanel.setVisible(false);
        
        return;
    }

    /**
     * フォント選択パネルの初期化
     */
    private void initFontChooser(){
        FontRenderContext oldContext = this.fontChooser.getFontRenderContext();
        boolean isAntiAliased         = oldContext.isAntiAliased();
        boolean usesFractionalMetrics = oldContext.usesFractionalMetrics();
        if(AppSetting.initfont != null){
            Font newFont = Font.decode(AppSetting.initfont);
            if( ! newFont.equals(this.fontChooser.getSelectedFont()) ){
                this.fontChooser.setSelectedFont(newFont);
            }
            if(FontChooser.guessBitmapFont(newFont)){
                isAntiAliased         = false;
                usesFractionalMetrics = false;
            }else{
                isAntiAliased         = true;
                usesFractionalMetrics = true;
            }
        }
        if(AppSetting.antialias != null){
            isAntiAliased = AppSetting.antialias.booleanValue();
        }
        if(AppSetting.fractional != null){
            usesFractionalMetrics = AppSetting.fractional.booleanValue();
        }
        FontRenderContext renderContext =
                new FontRenderContext(oldContext.getTransform(),
                                      isAntiAliased,
                                      usesFractionalMetrics );
        this.fontChooser.setFontRenderContext(renderContext);

        return;
    }
    
    /**
     * About画面を表示する。
     */
    private void actionAbout(){
        String message =
                Jindolf.TITLE
                + "   Version " + Jindolf.VERSION + "\n"
                + Jindolf.COPYRIGHT + "\n"
                + "ライセンス: " + Jindolf.LICENSE + "\n"
                + "連絡先: " + Jindolf.CONTACT;

        JOptionPane pane = new JOptionPane(message,
                                           JOptionPane.INFORMATION_MESSAGE,
                                           JOptionPane.DEFAULT_OPTION,
                                           GUIUtils.getLogoIcon());

        JDialog dialog = pane.createDialog(topFrame,
                                           Jindolf.TITLE + "について");

        dialog.pack();
        dialog.setVisible(true);
        dialog.dispose();

        return;
    }

    /**
     * アプリ終了。
     */
    private void actionExit(){
        Jindolf.exit(0);
    }

    /**
     * Help画面を表示する。
     */
    private void actionHelp(){
        if(helpFrame != null){                 // show Toggle
            toggleWindow(helpFrame);
            return;
        }

        helpFrame = new JFrame(Jindolf.TITLE + " ヘルプ");

        helpFrame.setResizable(true);
        Toolkit kit = helpFrame.getToolkit();
        kit.setDynamicLayout(false);
        helpFrame.setIconImage(GUIUtils.getWindowIconImage());
        helpFrame.setLocationByPlatform(true);

        JEditorPane edit = new JEditorPane();
        edit.setEditable(false);
        edit.setContentType("text/html");
        edit.putClientProperty(JEditorPane.W3C_LENGTH_UNITS, Boolean.TRUE);
        edit.setBorder(new EmptyBorder(0,0,0,0));

        URL htmlurl = Jindolf.getResource(HELP_HTML);
        try{
            edit.setPage(htmlurl);
        }catch(IOException e){
            Jindolf.logger.log(Level.WARNING,
                               "ヘルプファイルが読み込めません",
                               e);
            helpFrame = null;
            return;
        }

        JScrollPane sc = new JScrollPane(edit);

        Container content = helpFrame.getContentPane();
        content.add(sc);

        helpFrame.pack();
        helpFrame.setSize(450, 450);
        helpFrame.setVisible(true);

        return;
    }

    /**
     * L&Fの変更を行う。
     */
    // TODO Nimbus対応
    private void actionChangeLaF(){
        String className = actionManager.getSelectedLookAndFeel();

        LookAndFeel lnf;
        try{
            Class<?> lnfClass = Class.forName(className);
            lnf = (LookAndFeel)( lnfClass.newInstance() );
        }catch(Exception e){
            String message = "このLook&Feel["
                            + className
                            + "]を読み込む事ができません。";
            Jindolf.logger.log(Level.WARNING, message, e);
            JOptionPane.showMessageDialog(
                this.topFrame,
                message,
                "Look&Feel - " + Jindolf.TITLE,
                JOptionPane.WARNING_MESSAGE );
            return;
        }

        try{
            UIManager.setLookAndFeel(lnf);
        }catch(UnsupportedLookAndFeelException e){
            String message = "このLook&Feel["
                            + lnf.getName()
                            + "]はサポートされていません。";
            Jindolf.logger.log(Level.WARNING, message, e);
            JOptionPane.showMessageDialog(
                this.topFrame,
                message,
                "Look&Feel - " + Jindolf.TITLE,
                JOptionPane.WARNING_MESSAGE );
            return;
        }

        Jindolf.logger.info(
                "Look&Feelが["
                +lnf.getName()
                +"]に変更されました。");
        
        final Runnable updateUITask = new Runnable(){
            public void run(){
                SwingUtilities.updateComponentTreeUI(topFrame);
                topFrame.validate();

                if(helpFrame != null){
                    SwingUtilities.updateComponentTreeUI(helpFrame);
                    helpFrame.validate();
                }

                if(showlogFrame != null){
                    SwingUtilities.updateComponentTreeUI(showlogFrame);
                    showlogFrame.validate();
                }

                if(filterFrame != null){
                    SwingUtilities.updateComponentTreeUI(filterFrame);
                    filterFrame.validate();
                    filterFrame.pack();
                }

                if(accountFrame != null){
                    SwingUtilities.updateComponentTreeUI(accountFrame);
                    accountFrame.validate();
                    accountFrame.pack();
                }

                if(fontChooser != null){
                    SwingUtilities.updateComponentTreeUI(fontChooser);
                    fontChooser.validate();
                }
                
                if(findPanel != null){
                    SwingUtilities.updateComponentTreeUI(findPanel);
                    findPanel.validate();
                    findPanel.pack();
                }
                
                if(daySummaryPanel != null){
                    SwingUtilities.updateComponentTreeUI(daySummaryPanel);
                    daySummaryPanel.validate();
                }
                
                return;
            }
        };

        Executor executor = Executors.newCachedThreadPool();
        executor.execute(new Runnable(){
            public void run(){
                setBusy(true);
                updateStatusBar("Look&Feelを更新中…");
                try{
                    SwingUtilities.invokeAndWait(updateUITask);
                }catch(Exception e){
                    Jindolf.logger.log(
                            Level.WARNING,
                            "Look&Feelの更新に失敗しました。",
                            e );
                }finally{
                    updateStatusBar("Look&Feelが更新されました");
                    setBusy(false);
                }
                return;
            }
        });

        return;
    }

    /**
     * 発言フィルタ画面を表示する。
     */
    private void actionShowFilter(){
        toggleWindow(this.filterFrame);
        return;
    }

    /**
     * アカウント管理画面を表示する。
     */
    private void actionShowAccount(){
        if(this.accountFrame != null){                 // show Toggle
            toggleWindow(this.accountFrame);
            return;
        }

        this.accountFrame = new AccountPanel(this.topFrame, this.model);
        this.accountFrame.pack();
        this.accountFrame.setVisible(true);

        return;
    }

    /**
     * ログ表示画面を表示する。
     */
    private void actionShowLog(){
        toggleWindow(this.showlogFrame);
        return;
    }
    
    /**
     * 発言表示フォント選択画面を表示する。
     */
    private void actionFontSelect(){
        this.fontChooser.setVisible(true);
        if(this.fontChooser.isCanceled()) return;
        
        final Font newFont = this.fontChooser.getSelectedFont();
        final FontRenderContext newContext =
                this.fontChooser.getFontRenderContext();
        
        Executor executor = Executors.newCachedThreadPool();
        executor.execute(new Runnable(){
            public void run(){
                setBusy(true);
                updateStatusBar("発言表示フォントを変更中…");
                try{
                    TabBrowser tabBrowser = topFrame.getTabBrowser();
                    tabBrowser.setFontInfo(newFont, newContext);
                }catch(Exception e){
                    Jindolf.logger.log(
                            Level.WARNING,
                            "発言表示フォントの変更に失敗しました。",
                            e );
                }finally{
                    updateStatusBar(
                            "発言表示フォントが変更されました");
                    setBusy(false);
                }
                return;
            }
        });
        
        return;
    }
    
    /**
     * 検索パネルを表示する。
     */
    private void actionShowFind(){
        this.findPanel.setVisible(true);
        if(this.findPanel.isCanceled()){
            updateFindPanel();
            return;
        }
        if(this.findPanel.isBulkSearch()){
            bulkSearch();
        }else{
            regexSearch();
        }
        return;
    }
    
    /**
     * 検索処理。
     */
    private void regexSearch(){
        Discussion discussion = currentDiscussion();
        if(discussion == null) return;

        RegexPattern pattern = this.findPanel.getRegexPattern();
        discussion.setRegexPattern(pattern);
        int hits = discussion.highlightRegex();  // TODO 別スレッド化
        
        updateStatusBar("［" + hits + "］件ヒットしました");
        
        return;
    }

    /**
     * 一括検索処理
     */
    private void bulkSearch(){
        Executor executor = Executors.newCachedThreadPool();
        executor.execute(new Runnable(){
            public void run(){
                taskBulkSearch();
                return;
            }
        });
    }
    
    /**
     * 一括検索処理。ヘビータスク版
     */
    private void taskBulkSearch(){
        taskLoadAllPeriod();
        int totalhits = 0;
        RegexPattern pattern = findPanel.getRegexPattern();
        StringBuilder hitDesc = new StringBuilder();
        TabBrowser browser = topFrame.getTabBrowser();
        for(JdfScroller scroller : browser.getScrollerList()){
            Discussion discussion = scroller.getDiscussion();
            if(discussion == null) continue;
            discussion.setRegexPattern(pattern);
            int hits = discussion.highlightRegex();
            totalhits += hits;
            
            if(hits > 0){
                Period period = discussion.getPeriod();
                hitDesc.append(' ').append(period.getDay()).append("d:");
                hitDesc.append(hits).append("件");
            }
        }
        updateStatusBar(
                "［" + totalhits + "］件ヒットしました。"
                +hitDesc.toString());
        return;
    }
    
    /**
     * 検索パネルに現在選択中のPeriodを反映させる。
     */
    private void updateFindPanel(){
        Discussion discussion = currentDiscussion();
        if(discussion == null) return;
        RegexPattern pattern = discussion.getRegexPattern();
        this.findPanel.setRegexPattern(pattern);
        return;
    }
    
    /**
     * 発言集計パネルを表示
     */
    private void actionDaySummary(){
        Discussion discussion = currentDiscussion();
        if(discussion == null) return;
        
        Period period = discussion.getPeriod();
        if(period == null) return;
        
        if(this.daySummaryPanel == null){
            this.daySummaryPanel = new DaySummary(this.topFrame);
            this.daySummaryPanel.pack();
            this.daySummaryPanel.setSize(400, 500);
        }
        
        this.daySummaryPanel.summaryPeriod(period);
        this.daySummaryPanel.setVisible(true);
        
        return;
    }
    
    /**
     * 検索結果の次候補へジャンプ
     */
    private void actionSearchNext(){
        Discussion discussion = currentDiscussion();
        if(discussion == null) return;
        
        discussion.nextHotTarget();
        
        return;
    }
    
    /**
     * 検索結果の全候補へジャンプ
     */
    private void actionSearchPrev(){
        Discussion discussion = currentDiscussion();
        if(discussion == null) return;
        
        discussion.prevHotTarget();
        
        return;
    }
    
    /**
     * Period表示の強制再更新処理。
     */
    private void actionReloadPeriod(){
        updatePeriod(true);
        return;
    }
    
    /**
     * 全日程の一括ロード。
     */
    private void actionLoadAllPeriod(){
        Executor executor = Executors.newCachedThreadPool();
        executor.execute(new Runnable(){
            public void run(){
                taskLoadAllPeriod();
                return;
            }
        });
        
        return;
    }
    
    /**
     * 全日程の一括ロード。ヘビータスク版
     */
    private void taskLoadAllPeriod(){
        setBusy(true);
        updateStatusBar("一括読み込み開始");
        try{
            TabBrowser browser = topFrame.getTabBrowser();
            Village village = browser.getVillage();
            if(village == null) return;
            for(JdfScroller scroller : browser.getScrollerList()){
                Discussion discussion = scroller.getDiscussion();
                if(discussion == null) continue;
                Period period = discussion.getPeriod();
                if(period == null) continue;
                String message =
                        "" + period.getDay() + "日目の"
                        +"データを読み込んでいます";
                updateStatusBar(message);
                try{
                    period.updatePeriod();
                }catch(IOException e){
                    showNetworkError(village, e);
                    return;
                }
                discussion.showTopics();
            }
        }finally{
            updateStatusBar("一括読み込み完了");
            setBusy(false);
        }
        return;
    }
    
    /**
     * 村一覧の再読み込み
     */
    private void actionReloadVillageList(){
        JTree tree = this.topFrame.getTreeView();
        TreePath path = tree.getSelectionPath();
        if(path == null) return;

        Land land = null;
        for(int ct=0; ct<path.getPathCount(); ct++){
            Object obj = path.getPathComponent(ct);
            if(obj instanceof Land){
                land = (Land) obj;
                break;
            }
        }
        if(land == null) return;

        this.topFrame.showInitPanel();
        
        execReloadVillageList(land);
        
        return;
    }

    /**
     * 選択文字列をクリップボードにコピーする。
     */
    private void actionCopySelected(){
        Discussion discussion = currentDiscussion();
        if(discussion == null) return;
        
        CharSequence copied = discussion.copySelected();
        if(copied == null) return;
        
        copied = StringUtils.suppressString(copied);
        updateStatusBar(
                "[" + copied + "]をクリップボードにコピーしました");
        return;
    }
    
    /**
     * 一発言のみクリップボードにコピーする。
     */
    private void actionCopyTalk(){
        Discussion discussion = currentDiscussion();
        if(discussion == null) return;
        
        CharSequence copied = discussion.copyTalk();
        if(copied == null) return;
        
        copied = StringUtils.suppressString(copied);
        updateStatusBar(
                "[" + copied + "]をクリップボードにコピーしました");
        return;
    }
    
    /**
     * 指定した国の村一覧を読み込む。
     * @param land 国
     */
    private void execReloadVillageList(final Land land){
        final LandsTree treePanel = this.topFrame.getLandsTree();
        Executor executor = Executors.newCachedThreadPool();
        executor.execute(new Runnable(){
            public void run(){
                setBusy(true);
                updateStatusBar("村一覧を読み込み中…");
                try{
                    try{
                        model.loadVillageList(land);
                    }catch(IOException e){
                        showNetworkError(land, e);
                    }
                    treePanel.expandLand(land);
                }finally{
                    updateStatusBar("村一覧の読み込み完了");
                    setBusy(false);
                }
                return;
            }
        });
        return;
    }

    /**
     * Period表示の更新処理。
     * @param force trueならPeriodデータを強制再読み込み。
     */
    private void updatePeriod(final boolean force){
        final TabBrowser tabBrowser = this.topFrame.getTabBrowser();
        final Village village = tabBrowser.getVillage();
        if(village == null) return;
        this.topFrame.setFrameTitle(village.getVillageFullName());

        final JdfScroller scroller = currentJdfScroller();
        final Discussion discussion = currentDiscussion();
        if(discussion == null) return;
        discussion.setTopicFilter(filterFrame);
        final Period period = discussion.getPeriod();
        if(period == null) return;

        Executor executor = Executors.newCachedThreadPool();
        executor.execute(new Runnable(){
            public void run(){
                setBusy(true);
                try{
                    boolean wasHot = loadPeriod();

                    if(wasHot && ! period.isHot() ){
                        if(updatePeriodList() != true) return;
                    }

                    renderBrowser();
                }finally{
                    setBusy(false);
                }
                return;
            }

            private boolean loadPeriod(){
                updateStatusBar("1日分のデータを読み込んでいます…");
                boolean wasHot;
                try{
                    wasHot = period.isHot();
                    try{
                        period.loadPeriod(force);
                    }catch(IOException e){
                        showNetworkError(village, e);
                    }
                }finally{
                    updateStatusBar("1日分のデータを読み終わりました");
                }
                return wasHot;
            }

            private boolean updatePeriodList(){
                updateStatusBar("村情報を読み直しています…");
                try{
                    village.updatePeriodList();
                }catch(IOException e){
                    showNetworkError(village, e);
                    return false;
                }
                try{
                    SwingUtilities.invokeAndWait(new Runnable(){
                        public void run(){
                            tabBrowser.setVillage(village);
                            return;
                        }
                    });
                }catch(Exception e){
                    Jindolf.logger.log(Level.SEVERE,
                                       "タブ操作で致命的な障害が発生しました",
                                       e);
                }
                updateStatusBar("村情報を読み直しました…");
                return true;
            }
            
            private void renderBrowser(){
                updateStatusBar("レンダリング中…");
                try{
                    final int lastPos = scroller.getVerticalPosition();
                    try{
                        SwingUtilities.invokeAndWait(new Runnable(){
                            public void run(){
                                discussion.showTopics();
                                return;
                            }
                        });
                    }catch(Exception e){
                        Jindolf.logger
                               .log(Level.SEVERE,
                                    "ブラウザ表示で致命的な障害が発生しました",
                                    e);
                    }
                    SwingUtilities.invokeLater(new Runnable(){
                        public void run(){
                            scroller.setVerticalPosition(lastPos);
                        }
                    });
                }finally{
                    updateStatusBar("レンダリング完了");
                }
                return;
            }
        });

        return;
    }

    /**
     * 発言フィルタの操作による更新処理。
     */
    private void filterChanged(){
        final Discussion discussion = currentDiscussion();
        if(discussion == null) return;
        discussion.setTopicFilter(filterFrame);

        Executor executor = Executors.newCachedThreadPool();
        executor.execute(new Runnable(){
            public void run(){
                setBusy(true);
                updateStatusBar("フィルタリング中…");
                try{
                    discussion.filtering();
                }finally{
                    updateStatusBar("フィルタリング完了");
                    setBusy(false);
                }
                return;
            }
        });

        return;
    }

    /**
     * 現在選択中のPeriodを内包するJdfScrollerを返す。
     * @return JdfScroller
     */
    private JdfScroller currentJdfScroller(){
        TabBrowser tb = this.topFrame.getTabBrowser();
        JdfScroller result = tb.currentJdfScroller();
        return result;
    }
    
    /**
     * 現在選択中のPeriodを内包するJdfBrowserを返す。
     * @return JdfScroller
     */
    private Discussion currentDiscussion(){
        JdfScroller scroller = currentJdfScroller();
        if(scroller == null) return null;
        Discussion result = scroller.getDiscussion();
        return result;
    }
    
    /**
     * フレーム表示のトグル処理
     * @param window フレーム
     */
    // TODO アイコン化からの復帰もサポートしたい
    private void toggleWindow(Window window){
        if(window == null) return;
        if(window.isVisible()){
            window.setVisible(false);
            window.dispose();
        }else{
            window.setVisible(true);
        }
        return;
    }

    /**
     * ネットワークエラーを通知するモーダルダイアログを表示する。
     * OKボタンを押すまでこのメソッドは戻ってこない。
     * @param e ネットワークエラー
     */
    public void showNetworkError(Village village, IOException e){
        Land land = village.getParentLand();
        showNetworkError(land, e);
        return;
    }

    /**
     * ネットワークエラーを通知するモーダルダイアログを表示する。
     * OKボタンを押すまでこのメソッドは戻ってこない。
     * @param e ネットワークエラー
     */
    public void showNetworkError(Land land, IOException e){
        Jindolf.logger.log(Level.WARNING,
                           "ネットワークで障害が発生しました", e);

        ServerAccess server = land.getServerAccess();
        String message =
                land.getLandName()
                +"を運営するサーバとの間の通信で"
                +"何らかのトラブルが発生しました。\n"
                +"相手サーバのURLは [ " + server.getBaseURL() + " ] だよ。\n"
                +"Webブラウザでも遊べないか確認してみてね!\n";

        JOptionPane pane = new JOptionPane(message,
                                           JOptionPane.WARNING_MESSAGE,
                                           JOptionPane.DEFAULT_OPTION );

        JDialog dialog = pane.createDialog(topFrame,
                                           "通信異常発生 - " + Jindolf.TITLE);

        dialog.pack();
        dialog.setVisible(true);
        dialog.dispose();

        return;
    }

    /**
     * ツリーリストで何らかの要素（国、村）がクリックされたときの処理。
     * @param event イベント
     */
    public void valueChanged(TreeSelectionEvent event){
        TreePath path = event.getNewLeadSelectionPath();
        if(path == null) return;

        Object selObj = path.getLastPathComponent();
        
        if( selObj instanceof Land ){
            Land land = (Land)selObj;
            this.topFrame.showLandInfo(land);
            this.actionManager.appearVillage(false);
            this.actionManager.appearPeriod(false);
        }else if( selObj instanceof Village ){
            final Village village = (Village)selObj;
            
            Executor executor = Executors.newCachedThreadPool();
            executor.execute(new Runnable(){
                public void run(){
                    setBusy(true);
                    updateStatusBar("村情報を読み込み中…");

                    try{
                        village.updatePeriodList();
                    }catch(IOException e){
                        showNetworkError(village, e);
                        return;
                    }finally{
                        updateStatusBar("村情報の読み込み完了");
                        setBusy(false);
                    }

                    actionManager.appearVillage(true);
                    topFrame.showVillageInfo(village);
                    
                    return;
                }
            });
        }
        
        return;
    }

    /**
     * Periodがタブ選択されたときもしくは発言フィルタが操作されたときの処理。
     * @param event イベント
     */
    public void stateChanged(ChangeEvent event){
        Object source = event.getSource();

        if(source == filterFrame){
            filterChanged();
        }else if(source instanceof TabBrowser){
            updateFindPanel();
            updatePeriod(false);
            Discussion discussion = currentDiscussion();
            if(discussion == null) this.actionManager.appearPeriod(false);
            else                   this.actionManager.appearPeriod(true);
        }
        return;
    }

    /**
     * アクションイベント処理。
     * 主にメニュー選択やボタン押下など。
     * @param e イベント
     */
    public void actionPerformed(ActionEvent e){
        if(this.isBusyNow) return;
        
        String cmd = e.getActionCommand();
        if      (cmd.equals(ActionManager.COMMAND_ABOUT)){
            actionAbout();
        }else if(cmd.equals(ActionManager.COMMAND_EXIT)){
            actionExit();
        }else if(cmd.equals(ActionManager.COMMAND_HELPDOC)){
            actionHelp();
        }else if(cmd.equals(ActionManager.COMMAND_LANDF)){
            actionChangeLaF();
        }else if(cmd.equals(ActionManager.COMMAND_FILTER)){
            actionShowFilter();
        }else if(cmd.equals(ActionManager.COMMAND_ACCOUNT)){
            actionShowAccount();
        }else if(cmd.equals(ActionManager.COMMAND_SHOWLOG)){
            actionShowLog();
        }else if(cmd.equals(ActionManager.COMMAND_FONTSELECT)){
            actionFontSelect();
        }else if(cmd.equals(ActionManager.COMMAND_RELOAD)){
            actionReloadPeriod();
        }else if(cmd.equals(ActionManager.COMMAND_VILLAGELIST)){
            actionReloadVillageList();
        }else if(cmd.equals(ActionManager.COMMAND_COPY)){
            actionCopySelected();
        }else if(cmd.equals(ActionManager.COMMAND_COPYTALK)){
            actionCopyTalk();
        }else if(cmd.equals(ActionManager.COMMAND_SEARCHNEXT)){
            actionSearchNext();
        }else if(cmd.equals(ActionManager.COMMAND_SEARCHPREV)){
            actionSearchPrev();
        }else if(cmd.equals(ActionManager.COMMAND_SHOWFIND)){
            actionShowFind();
        }else if(cmd.equals(ActionManager.COMMAND_ALLPERIOD)){
            actionLoadAllPeriod();
        }else if(cmd.equals(ActionManager.COMMAND_DAYSUMMARY)){
            actionDaySummary();
        }
        return;
    }

    /**
     * 村選択ツリーリストが畳まれるとき呼ばれる。
     * @param event ツリーイベント
     */
    public void treeWillCollapse(TreeExpansionEvent event){
        return;
    }

    /**
     * 村選択ツリーリストが展開されるとき呼ばれる。
     * @param event ツリーイベント
     */
    public void treeWillExpand(TreeExpansionEvent event){
        if(!(event.getSource() instanceof JTree)){
            return;
        }

        TreePath path = event.getPath();
        Object lastObj = path.getLastPathComponent();
        if(!(lastObj instanceof Land)){
            return;
        }
        final Land land = (Land) lastObj;
        if(land.getVillageCount() > 0){
            return;
        }

        execReloadVillageList(land);

        return;
    }
    
    /**
     * アンカー展開処理
     * @param event アンカーヒットイベント
     */
    public void anchorHitted(AnchorHitEvent event){
        final Discussion discussion = currentDiscussion();
        if(discussion == null) return;
        Period period = discussion.getPeriod();
        if(period == null) return;
        final Village village = period.getVillage();

        final TalkDraw talkDraw = event.getTalkDraw();
        final Anchor anchor = event.getAnchor();

        Executor executor = Executors.newCachedThreadPool();
        executor.execute(new Runnable(){
            public void run(){
                setBusy(true);
                updateStatusBar("アンカーの展開中…");

                final List<Talk> talkList;
                try{
                    talkList = village.getTalkListFromAnchor(anchor);
                    if(talkList.size() <= 0){
                        updateStatusBar(
                                  "アンカーの展開先["
                                + anchor.toString()
                                + "]が見つかりません");
                        return;
                    }
                    SwingUtilities.invokeLater(new Runnable(){
                        public void run(){
                            talkDraw.showAnchorTalks(anchor, talkList);
                            discussion.layoutRows();
                            return;
                        }
                    });
                    updateStatusBar(
                              "アンカー["
                            + anchor.toString()
                            + "]の展開完了");
                }catch(IOException e){
                    updateStatusBar(
                            "アンカーの展開中にエラーが起きました");
                }finally{
                    setBusy(false);
                }
                
                return;
            }
        });
        
        return;
    }
    
    /**
     * ヘビーなタスク実行をアピール
     * @param isBusy ヘビーならtrue
     */
    private void setBusy(boolean isBusy){
        this.isBusyNow = isBusy;
        this.topFrame.setBusy(isBusy);
    }
    
    /**
     * ステータスバーを更新する
     * @param message メッセージ
     */
    private void updateStatusBar(String message){
        this.topFrame.updateSysMessage(message);
    }
}
